@@ -1,12 +1,15 @@
-# This file is Copyright (c) 2013-2014 Sebastien Bourdeauducq <sb@m-labs.hk>
-# This file is Copyright (c) 2013-2020 Florent Kermarrec <florent@enjoy-digital.fr>
-# This file is Copyright (c) 2017 whitequark <whitequark@whitequark.org>
-# This file is Copyright (c) 2014 Yann Sionneau <ys@m-labs.hk>
-# This file is Copyright (c) 2018 bunnie <bunnie@kosagi.com>
-# This file is Copyright (c) 2019 Gabriel L. Somlo <gsomlo@gmail.com>
-# License: BSD
+#
+# This file is part of LiteDRAM.
+#
+# Copyright (c) 2013-2014 Sebastien Bourdeauducq <sb@m-labs.hk>
+# Copyright (c) 2013-2020 Florent Kermarrec <florent@enjoy-digital.fr>
+# Copyright (c) 2017 whitequark <whitequark@whitequark.org>
+# Copyright (c) 2014 Yann Sionneau <ys@m-labs.hk>
+# Copyright (c) 2018 bunnie <bunnie@kosagi.com>
+# Copyright (c) 2019 Gabriel L. Somlo <gsomlo@gmail.com>
+# SPDX-License-Identifier: BSD-2-Clause
 
-from migen import log2_int
+from migen import *
 
 cmds = {
     "PRECHARGE_ALL": "DFII_COMMAND_RAS|DFII_COMMAND_WE|DFII_COMMAND_CS",
@@ -352,6 +355,8 @@ def get_ddr4_phy_init_sequence(phy_settings, timing_settings):
     rtt_wr  = "120ohm"
     ron     = "34ohm"
     tdqs    = 0
+    dm      = 1
+    assert not (dm and tdqs)
 
     # override electrical settings if specified
     if hasattr(phy_settings, "rtt_nom"):
@@ -369,7 +374,7 @@ def get_ddr4_phy_init_sequence(phy_settings, timing_settings):
     mr2 = format_mr2(cwl, z_to_rtt_wr[rtt_wr])
     mr3 = format_mr3(timing_settings.fine_refresh_mode)
     mr4 = 0
-    mr5 = 0
+    mr5 = (dm << 10)
     mr6 = format_mr6(4) # FIXME: tCCD
 
     rdimm_init = []
@@ -395,7 +400,7 @@ def get_ddr4_phy_init_sequence(phy_settings, timing_settings):
         def get_fine_speed(tck):
             # JESD82-31A page 83
             freq = 2/tck
-            fine_speed = (freq - 1240e6) // 20e6
+            fine_speed = int((freq - 1240e6) // 20e6)
             fine_speed = max(fine_speed, 0)
             fine_speed = min(fine_speed, 0b1100001)
             return fine_speed
@@ -482,24 +487,43 @@ def get_sdram_phy_c_header(phy_settings, timing_settings):
     r += "#define SDRAM_PHY_XDR "+str(1 if phy_settings.memtype == "SDR" else 2) + "\n"
     r += "#define SDRAM_PHY_DATABITS "+str(phy_settings.databits) + "\n"
     r += "#define SDRAM_PHY_PHASES "+str(nphases)+"\n"
+    if phy_settings.cl is not None:
+        r += "#define SDRAM_PHY_CL "+str(phy_settings.cl)+"\n"
+    if phy_settings.cwl is not None:
+        r += "#define SDRAM_PHY_CWL "+str(phy_settings.cwl)+"\n"
+    if phy_settings.cmd_latency is not None:
+        r += "#define SDRAM_PHY_CMD_LATENCY "+str(phy_settings.cmd_latency)+"\n"
+    if phy_settings.cmd_delay is not None:
+        r += "#define SDRAM_PHY_CMD_DELAY "+str(phy_settings.cmd_delay)+"\n"
+
+    # Define PHY Read.Write phases
+    rdphase = phy_settings.rdphase
+    if isinstance(rdphase, Signal): rdphase = rdphase.reset.value
+    r += "#define SDRAM_PHY_RDPHASE "+str(rdphase)+"\n"
+    wrphase = phy_settings.wrphase
+    if isinstance(wrphase, Signal): wrphase = wrphase.reset.value
+    r += "#define SDRAM_PHY_WRPHASE "+str(wrphase)+"\n"
 
     # Define Read/Write Leveling capability
     if phytype in ["USDDRPHY", "USPDDRPHY", "K7DDRPHY", "V7DDRPHY"]:
         r += "#define SDRAM_PHY_WRITE_LEVELING_CAPABLE\n"
     if phytype in ["USDDRPHY", "USPDDRPHY"]:
         r += "#define SDRAM_PHY_WRITE_LEVELING_REINIT\n"
-    if phytype in ["USDDRPHY", "USPDDRPHY", "A7DDRPHY", "K7DDRPHY", "V7DDRPHY", "ECP5DDRPHY"]:
+    if phytype in ["USDDRPHY", "USPDDRPHY", "A7DDRPHY", "K7DDRPHY", "V7DDRPHY"]:
+        r += "#define SDRAM_PHY_WRITE_LATENCY_CALIBRATION_CAPABLE\n"
+        r += "#define SDRAM_PHY_READ_LEVELING_CAPABLE\n"
+    if phytype in ["ECP5DDRPHY"]:
         r += "#define SDRAM_PHY_READ_LEVELING_CAPABLE\n"
 
     # Define number of modules/delays/bitslips
     if phytype in ["USDDRPHY", "USPDDRPHY"]:
         r += "#define SDRAM_PHY_MODULES DFII_PIX_DATA_BYTES/2\n"
         r += "#define SDRAM_PHY_DELAYS 512\n"
-        r += "#define SDRAM_PHY_BITSLIPS 16\n"
+        r += "#define SDRAM_PHY_BITSLIPS 8\n"
     elif phytype in ["A7DDRPHY", "K7DDRPHY", "V7DDRPHY"]:
         r += "#define SDRAM_PHY_MODULES DFII_PIX_DATA_BYTES/2\n"
         r += "#define SDRAM_PHY_DELAYS 32\n"
-        r += "#define SDRAM_PHY_BITSLIPS 16\n"
+        r += "#define SDRAM_PHY_BITSLIPS 8\n"
     elif phytype in ["ECP5DDRPHY"]:
         r += "#define SDRAM_PHY_MODULES DFII_PIX_DATA_BYTES/4\n"
         r += "#define SDRAM_PHY_DELAYS 8\n"
@@ -513,7 +537,7 @@ def get_sdram_phy_c_header(phy_settings, timing_settings):
 
     r += "static void cdelay(int i);\n"
 
-    # commands_px functions
+    # Commands functions
     for n in range(nphases):
         r += """
 __attribute__((unused)) static void command_p{n}(int cmd)
@@ -523,20 +547,7 @@ __attribute__((unused)) static void command_p{n}(int cmd)
 }}""".format(n=str(n))
     r += "\n\n"
 
-    # rd/wr access macros
-    r += """
-#define sdram_dfii_pird_address_write(X) sdram_dfii_pi{rdphase}_address_write(X)
-#define sdram_dfii_piwr_address_write(X) sdram_dfii_pi{wrphase}_address_write(X)
-#define sdram_dfii_pird_baddress_write(X) sdram_dfii_pi{rdphase}_baddress_write(X)
-#define sdram_dfii_piwr_baddress_write(X) sdram_dfii_pi{wrphase}_baddress_write(X)
-#define command_prd(X) command_p{rdphase}(X)
-#define command_pwr(X) command_p{wrphase}(X)
-""".format(rdphase=str(phy_settings.rdphase), wrphase=str(phy_settings.wrphase))
-    r += "\n"
-
-    #
-    # sdrrd/sdrwr functions utilities
-    #
+    # Write/Read functions
     r += "#define DFII_PIX_DATA_SIZE CSR_SDRAM_DFII_PI0_WRDATA_SIZE\n"
     sdram_dfii_pix_wrdata_addr = []
     for n in range(nphases):
@@ -560,7 +571,7 @@ const unsigned long sdram_dfii_pix_rddata_addr[SDRAM_PHY_PHASES] = {{
     init_sequence, mr1 = get_sdram_phy_init_sequence(phy_settings, timing_settings)
 
     if phy_settings.memtype in ["DDR3", "DDR4"]:
-        # the value of MR1 needs to be modified during write leveling
+        # The value of MR1 needs to be modified during write leveling
         r += "#define DDRX_MR1 {}\n\n".format(mr1)
 
     r += "static void init_sequence(void)\n{\n"
