@@ -1,16 +1,19 @@
-# This file is Copyright (c) 2015 Sebastien Bourdeauducq <sb@m-labs.hk>
-# This file is Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
-# This file is Copyright (c) 2018 John Sully <john@csquare.ca>
-# This file is Copyright (c) 2019 Ambroz Bizjak <abizjak.pro@gmail.com>
-# This file is Copyright (c) 2019 Antony Pavlov <antonynpavlov@gmail.com>
-# This file is Copyright (c) 2018 bunnie <bunnie@kosagi.com>
-# This file is Copyright (c) 2018 David Shah <dave@ds0.me>
-# This file is Copyright (c) 2019 Steve Haynal - VSD Engineering
-# This file is Copyright (c) 2018 Tim 'mithro' Ansell <me@mith.ro>
-# This file is Copyright (c) 2018 Daniel Kucera <daniel.kucera@gmail.com>
-# This file is Copyright (c) 2018 Mikołaj Sowiński <mikolaj.sowinski@gmail.com>
-# This file is Copyright (c) 2020 Antmicro <www.antmicro.com>
-# License: BSD
+#
+# This file is part of LiteDRAM.
+#
+# Copyright (c) 2015 Sebastien Bourdeauducq <sb@m-labs.hk>
+# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
+# Copyright (c) 2018 John Sully <john@csquare.ca>
+# Copyright (c) 2019 Ambroz Bizjak <abizjak.pro@gmail.com>
+# Copyright (c) 2019 Antony Pavlov <antonynpavlov@gmail.com>
+# Copyright (c) 2018 bunnie <bunnie@kosagi.com>
+# Copyright (c) 2018 David Shah <dave@ds0.me>
+# Copyright (c) 2019 Steve Haynal - VSD Engineering
+# Copyright (c) 2018 Tim 'mithro' Ansell <me@mith.ro>
+# Copyright (c) 2018 Daniel Kucera <daniel.kucera@gmail.com>
+# Copyright (c) 2018 Mikołaj Sowiński <mikolaj.sowinski@gmail.com>
+# Copyright (c) 2020 Antmicro <www.antmicro.com>
+# SPDX-License-Identifier: BSD-2-Clause
 
 from math import ceil
 from collections import namedtuple
@@ -508,8 +511,7 @@ class M12L16161A(SDRModule):
 class DDRModule(SDRAMModule):                     memtype = "DDR"
 class DDRRegisteredModule(SDRAMRegisteredModule): memtype = "DDR"
 
-class MT46V32M16(SDRAMModule):
-    memtype = "DDR"
+class MT46V32M16(DDRModule):
     # geometry
     nbanks = 4
     nrows  = 8192
@@ -547,7 +549,6 @@ class DDR2Module(SDRAMModule):                     memtype = "DDR2"
 class DDR2RegisteredModule(SDRAMRegisteredModule): memtype = "DDR2"
 
 class MT47H128M8(DDR2Module):
-    memtype = "DDR2"
     # geometry
     nbanks = 8
     nrows  = 16384
@@ -557,7 +558,6 @@ class MT47H128M8(DDR2Module):
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 127.5), tFAW=None, tRAS=None)}
 
 class MT47H32M16(DDR2Module):
-    memtype = "DDR2"
     # geometry
     nbanks = 4
     nrows  = 8192
@@ -567,7 +567,6 @@ class MT47H32M16(DDR2Module):
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 127.5), tFAW=None, tRAS=None)}
 
 class MT47H64M16(DDR2Module):
-    memtype = "DDR2"
     # geometry
     nbanks = 8
     nrows  = 8192
@@ -577,7 +576,6 @@ class MT47H64M16(DDR2Module):
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 127.5), tFAW=None, tRAS=None)}
 
 class P3R1GE4JGF(DDR2Module):
-    memtype = "DDR2"
     # geometry
     nbanks = 8
     nrows  = 8192
@@ -592,7 +590,6 @@ class DDR3Module(SDRAMModule):                     memtype = "DDR3"
 class DDR3RegisteredModule(SDRAMRegisteredModule): memtype = "DDR3"
 
 class MT41K64M16(DDR3Module):
-    memtype = "DDR3"
     # geometry
     nbanks = 8
     nrows  = 8192
@@ -608,7 +605,6 @@ class MT41K64M16(DDR3Module):
     speedgrade_timings["default"] = speedgrade_timings["1600"]
 
 class MT41J128M16(DDR3Module):
-    memtype = "DDR3"
     # geometry
     nbanks = 8
     nrows  = 16384
