@@ -1,11 +1,17 @@
-# This file is Copyright (c) 2015-2020 Florent Kermarrec <florent@enjoy-digital.fr>
-# This file is Copyright (c) 2015 Sebastien Bourdeauducq <sb@m-labs.hk>
-# License: BSD
+#
+# This file is part of LiteDRAM.
+#
+# Copyright (c) 2015-2020 Florent Kermarrec <florent@enjoy-digital.fr>
+# Copyright (c) 2015 Sebastien Bourdeauducq <sb@m-labs.hk>
+# SPDX-License-Identifier: BSD-2-Clause
 
 # 1:4, 1:2 frequency-ratio DDR2/DDR3 PHY for Xilinx's Series7
 # DDR2: 400, 533, 667, 800 and 1066 MT/s
 # DDR3: 800, 1066, 1333 and 1600 MT/s
 
+from functools import reduce
+from operator import or_
+
 import math
 
 from migen import *
@@ -24,10 +30,8 @@ class S7DDRPHY(Module, AutoCSR):
         sys_clk_freq     = 100e6,
         iodelay_clk_freq = 200e6,
         cmd_latency      = 0,
-        interface_type   = "NETWORKING"):
+        cmd_delay        = None):
         assert not (memtype == "DDR3" and nphases == 2)
-        assert interface_type in ["NETWORKING", "MEMORY"]
-        assert not (interface_type == "MEMORY" and nphases == 2)
         phytype     = self.__class__.__name__
         pads        = PHYPadsCombiner(pads)
         tck         = 2/(2*nphases*sys_clk_freq)
@@ -48,18 +52,22 @@ class S7DDRPHY(Module, AutoCSR):
 
         cl, cwl         = get_cl_cw(memtype, tck)
         cl_sys_latency  = get_sys_latency(nphases, cl)
-        cwl             = cwl + cmd_latency
         cwl_sys_latency = get_sys_latency(nphases, cwl)
+        rdphase         = get_sys_phase(nphases, cl_sys_latency,   cl + cmd_latency)
+        wrphase         = get_sys_phase(nphases, cwl_sys_latency, cwl + cmd_latency)
 
         # Registers --------------------------------------------------------------------------------
+        self._rst             = CSRStorage()
+
         self._dly_sel         = CSRStorage(databits//8)
         self._half_sys8x_taps = CSRStorage(5, reset=half_sys8x_taps)
 
         self._wlevel_en     = CSRStorage()
         self._wlevel_strobe = CSR()
 
-        self._cdly_rst = CSR()
-        self._cdly_inc = CSR()
+        if with_odelay:
+            self._cdly_rst = CSR()
+            self._cdly_inc = CSR()
 
         self._dly_sel = CSRStorage(databits//8)
 
@@ -74,13 +82,13 @@ class S7DDRPHY(Module, AutoCSR):
             self._wdly_dqs_rst  = CSR()
             self._wdly_dqs_inc  = CSR()
 
+        self._wdly_dq_bitslip_rst = CSR()
+        self._wdly_dq_bitslip     = CSR()
+
+        self._rdphase = CSRStorage(int(math.log2(nphases)), reset=rdphase)
+        self._wrphase = CSRStorage(int(math.log2(nphases)), reset=wrphase)
+
         # PHY settings -----------------------------------------------------------------------------
-        rdcmdphase, rdphase = get_sys_phases(nphases, cl_sys_latency, cl)
-        wrcmdphase, wrphase = get_sys_phases(nphases, cwl_sys_latency, cwl)
-        iserdese2_latency  = {
-            "NETWORKING": 2,
-            "MEMORY":     1,
-        }
         self.settings = PhySettings(
             phytype       = phytype,
             memtype       = memtype,
@@ -88,14 +96,14 @@ class S7DDRPHY(Module, AutoCSR):
             dfi_databits  = 2*databits,
             nranks        = nranks,
             nphases       = nphases,
-            rdphase       = rdphase,
-            wrphase       = wrphase,
-            rdcmdphase    = rdcmdphase,
-            wrcmdphase    = wrcmdphase,
+            rdphase       = self._rdphase.storage,
+            wrphase       = self._wrphase.storage,
             cl            = cl,
-            cwl           = cwl - cmd_latency,
-            read_latency  = 2 + cl_sys_latency + iserdese2_latency[interface_type] + 2,
-            write_latency = cwl_sys_latency,
+            cwl           = cwl,
+            read_latency  = cl_sys_latency + 6,
+            write_latency = cwl_sys_latency - 1,
+            cmd_latency   = cmd_latency,
+            cmd_delay     = cmd_delay,
         )
 
         # DFI Interface ----------------------------------------------------------------------------
@@ -118,17 +126,10 @@ class S7DDRPHY(Module, AutoCSR):
                     p_TRISTATE_WIDTH = 1,
                     p_DATA_RATE_OQ   = "DDR",
                     p_DATA_RATE_TQ   = "BUF",
-                    i_RST    = ResetSignal(),
+                    i_RST    = ResetSignal() | self._rst.storage,
                     i_CLK    = ClockSignal(ddr_clk),
                     i_CLKDIV = ClockSignal(),
-                    i_D1     = 0,
-                    i_D2     = 1,
-                    i_D3     = 0,
-                    i_D4     = 1,
-                    i_D5     = 0,
-                    i_D6     = 1,
-                    i_D7     = 0,
-                    i_D8     = 1,
+                    **{f"i_D{n+1}": (0b10101010 >> n) & 0b1 for n in range(8)},
                     o_OQ     = sd_clk_se_nodelay,
                     i_OCE    = 1,
                 )
@@ -143,7 +144,7 @@ class S7DDRPHY(Module, AutoCSR):
                         p_ODELAY_TYPE           = "VARIABLE",
                         p_ODELAY_VALUE          = 0,
                         i_C        = ClockSignal(),
-                        i_LD       = self._cdly_rst.re,
+                        i_LD       = self._cdly_rst.re | self._rst.storage,
                         i_LDPIPEEN = 0,
                         i_CE       = self._cdly_inc.re,
                         i_INC      = 1,
@@ -156,114 +157,34 @@ class S7DDRPHY(Module, AutoCSR):
                     o_OB = pads.clk_n[i]
                 )
 
-            # Addresses and Commands -------------------------------------------------------------------
-            for i in range(addressbits):
-                address = Signal()
-                self.specials += Instance("OSERDESE2",
-                    p_SERDES_MODE    = "MASTER",
-                    p_DATA_WIDTH     = 2*nphases,
-                    p_TRISTATE_WIDTH = 1,
-                    p_DATA_RATE_OQ   = "DDR",
-                    p_DATA_RATE_TQ   = "BUF",
-                    i_RST    = ResetSignal(),
-                    i_CLK    = ClockSignal(ddr_clk),
-                    i_CLKDIV = ClockSignal(),
-                    i_D1     = dfi.phases[0].address[i],
-                    i_D2     = dfi.phases[0].address[i],
-                    i_D3     = dfi.phases[1].address[i],
-                    i_D4     = dfi.phases[1].address[i],
-                    i_D5     = dfi.phases[2].address[i],
-                    i_D6     = dfi.phases[2].address[i],
-                    i_D7     = dfi.phases[3].address[i],
-                    i_D8     = dfi.phases[3].address[i],
-                    i_OCE    = 1,
-                    o_OQ     = address if with_odelay else pads.a[i],
-                )
-                if with_odelay:
-                    self.specials += Instance("ODELAYE2",
-                        p_SIGNAL_PATTERN        = "DATA",
-                        p_DELAY_SRC             = "ODATAIN",
-                        p_CINVCTRL_SEL          = "FALSE",
-                        p_HIGH_PERFORMANCE_MODE = "TRUE",
-                        p_PIPE_SEL              = "FALSE",
-                        p_REFCLK_FREQUENCY      = iodelay_clk_freq/1e6,
-                        p_ODELAY_TYPE           = "VARIABLE",
-                        p_ODELAY_VALUE          = 0,
-                        i_C        = ClockSignal(),
-                        i_LD       = self._cdly_rst.re,
-                        i_LDPIPEEN = 0,
-                        i_CE       = self._cdly_inc.re,
-                        i_INC      = 1,
-                        o_ODATAIN  = address,
-                        o_DATAOUT  = pads.a[i],
-                    )
-            for i in range(bankbits):
-                bank = Signal()
-                self.specials += Instance("OSERDESE2",
-                    p_SERDES_MODE    = "MASTER",
-                    p_DATA_WIDTH     = 2*nphases,
-                    p_TRISTATE_WIDTH = 1,
-                    p_DATA_RATE_OQ   = "DDR",
-                    p_DATA_RATE_TQ   = "BUF",
-                    i_RST    = ResetSignal(),
-                    i_CLK    = ClockSignal(ddr_clk),
-                    i_CLKDIV = ClockSignal(),
-                    i_D1     = dfi.phases[0].bank[i],
-                    i_D2     = dfi.phases[0].bank[i],
-                    i_D3     = dfi.phases[1].bank[i],
-                    i_D4     = dfi.phases[1].bank[i],
-                    i_D5     = dfi.phases[2].bank[i],
-                    i_D6     = dfi.phases[2].bank[i],
-                    i_D7     = dfi.phases[3].bank[i],
-                    i_D8     = dfi.phases[3].bank[i],
-                    i_OCE    = 1,
-                    o_OQ     = bank if with_odelay else pads.ba[i],
-                )
-                if with_odelay:
-                    self.specials += Instance("ODELAYE2",
-                        p_SIGNAL_PATTERN        = "DATA",
-                        p_DELAY_SRC             = "ODATAIN",
-                        p_CINVCTRL_SEL          = "FALSE",
-                        p_HIGH_PERFORMANCE_MODE = "TRUE",
-                        p_PIPE_SEL              = "FALSE",
-                        p_REFCLK_FREQUENCY      = iodelay_clk_freq/1e6,
-                        p_ODELAY_TYPE           = "VARIABLE",
-                        p_ODELAY_VALUE          = 0,
-                        i_C        = ClockSignal(),
-                        i_LD       = self._cdly_rst.re,
-                        i_LDPIPEEN = 0,
-                        i_CE       = self._cdly_inc.re,
-                        i_INC      = 1,
-                        o_ODATAIN  = bank,
-                        o_DATAOUT  = pads.ba[i],
-                    )
-            controls = ["ras_n", "cas_n", "we_n", "cke", "odt"]
-            if hasattr(pads, "reset_n"):
-                controls.append("reset_n")
-            if hasattr(pads, "cs_n"):
-                controls.append("cs_n")
-            for name in controls:
-                for i in range(len(getattr(pads, name))):
-                    cmd = Signal()
+            # Commands -----------------------------------------------------------------------------
+            commands = {
+                "a"    : "address",
+                "ba"   : "bank"   ,
+                "ras_n": "ras_n"  ,
+                "cas_n": "cas_n"  ,
+                "we_n" : "we_n"   ,
+                "cke"  : "cke"    ,
+                "odt"  : "odt"    ,
+            }
+            if hasattr(pads, "reset_n"): commands.update({"reset_n" : "reset_n"})
+            if hasattr(pads, "cs_n")   : commands.update({"cs_n"    : "cs_n"})
+            for pad_name, dfi_name in commands.items():
+                pad = getattr(pads, pad_name)
+                for i in range(len(pad)):
+                    oq  = Signal()
                     self.specials += Instance("OSERDESE2",
                         p_SERDES_MODE    = "MASTER",
                         p_DATA_WIDTH     = 2*nphases,
                         p_TRISTATE_WIDTH = 1,
                         p_DATA_RATE_OQ   = "DDR",
                         p_DATA_RATE_TQ   = "BUF",
-                        i_RST    = ResetSignal(),
+                        i_RST    = ResetSignal() | self._rst.storage,
                         i_CLK    = ClockSignal(ddr_clk),
                         i_CLKDIV = ClockSignal(),
-                        i_D1     = getattr(dfi.phases[0], name)[i],
-                        i_D2     = getattr(dfi.phases[0], name)[i],
-                        i_D3     = getattr(dfi.phases[1], name)[i],
-                        i_D4     = getattr(dfi.phases[1], name)[i],
-                        i_D5     = getattr(dfi.phases[2], name)[i],
-                        i_D6     = getattr(dfi.phases[2], name)[i],
-                        i_D7     = getattr(dfi.phases[3], name)[i],
-                        i_D8     = getattr(dfi.phases[3], name)[i],
+                        **{f"i_D{n+1}": getattr(dfi.phases[n//2], dfi_name)[i] for n in range(8)},
                         i_OCE    = 1,
-                        o_OQ     = cmd if with_odelay else getattr(pads, name)[i],
+                        o_OQ     = oq if with_odelay else pad[i],
                     )
                     if with_odelay:
                         self.specials += Instance("ODELAYE2",
@@ -276,92 +197,52 @@ class S7DDRPHY(Module, AutoCSR):
                             p_ODELAY_TYPE           = "VARIABLE",
                             p_ODELAY_VALUE          = 0,
                             i_C        = ClockSignal(),
-                            i_LD       = self._cdly_rst.re,
+                            i_LD       = self._cdly_rst.re | self._rst.storage,
                             i_LDPIPEEN = 0,
                             i_CE       = self._cdly_inc.re,
                             i_INC      = 1,
-                            o_ODATAIN  = cmd,
-                            o_DATAOUT  = getattr(pads, name)[i],
+                            o_ODATAIN  = oq,
+                            o_DATAOUT  = pad[i],
                         )
 
-        # DQS and DM -------------------------------------------------------------------------------
-        dqs_oe         = Signal()
-        dqs_oe_delayed = Signal() # Tristate control is asynchronous, needs to be delayed.
-        dqs_pattern    = DQSPattern(
+        # DQS --------------------------------------------------------------------------------------
+        dqs_oe        = Signal()
+        dqs_preamble  = Signal()
+        dqs_postamble = Signal()
+        dqs_oe_delay  = TappedDelayLine(ntaps=2 if nphases == 4 else 1)
+        dqs_pattern   = DQSPattern(
+            #preamble      = dqs_preamble,  # FIXME
+            #postamble     = dqs_postamble, # FIXME
             wlevel_en     = self._wlevel_en.storage,
             wlevel_strobe = self._wlevel_strobe.re,
             register      = not with_odelay)
-        self.submodules += dqs_pattern
-        self.sync += dqs_oe_delayed.eq(dqs_pattern.preamble | dqs_oe | dqs_pattern.postamble)
-        for i in range(databits//8):
-            dm_o_nodelay = Signal()
-            self.specials += Instance("OSERDESE2",
-                p_SERDES_MODE    = "MASTER",
-                p_DATA_WIDTH     = 2*nphases,
-                p_TRISTATE_WIDTH = 1,
-                p_DATA_RATE_OQ   = "DDR",
-                p_DATA_RATE_TQ   = "BUF",
-                i_RST    = ResetSignal(),
-                i_CLK    = ClockSignal(ddr_clk),
-                i_CLKDIV = ClockSignal(),
-                i_D1     = dfi.phases[0].wrdata_mask[i],
-                i_D2     = dfi.phases[0].wrdata_mask[databits//8+i],
-                i_D3     = dfi.phases[1].wrdata_mask[i],
-                i_D4     = dfi.phases[1].wrdata_mask[databits//8+i],
-                i_D5     = dfi.phases[2].wrdata_mask[i],
-                i_D6     = dfi.phases[2].wrdata_mask[databits//8+i],
-                i_D7     = dfi.phases[3].wrdata_mask[i],
-                i_D8     = dfi.phases[3].wrdata_mask[databits//8+i],
-                i_OCE    = 1,
-                o_OQ     = dm_o_nodelay if with_odelay else pads.dm[i],
-            )
-            if with_odelay:
-                self.specials += Instance("ODELAYE2",
-                    p_SIGNAL_PATTERN        = "DATA",
-                    p_DELAY_SRC             = "ODATAIN",
-                    p_CINVCTRL_SEL          = "FALSE",
-                    p_HIGH_PERFORMANCE_MODE = "TRUE",
-                    p_PIPE_SEL              = "FALSE",
-                    p_REFCLK_FREQUENCY      = iodelay_clk_freq/1e6,
-                    p_ODELAY_TYPE           = "VARIABLE",
-                    p_ODELAY_VALUE          = 0,
-                    i_C        = ClockSignal(),
-                    i_LD       = self._dly_sel.storage[i] & self._wdly_dq_rst.re,
-                    i_LDPIPEEN = 0,
-                    i_CE       = self._dly_sel.storage[i] & self._wdly_dq_inc.re,
-                    i_INC      = 1,
-                    o_ODATAIN  = dm_o_nodelay,
-                    o_DATAOUT  = pads.dm[i],
-                )
-
-        dqs_i         = Signal(databits//8)
-        dqs_i_delayed = Signal(databits//8)
+        self.submodules += dqs_oe_delay, dqs_pattern
+        self.comb += dqs_oe_delay.input.eq(dqs_preamble | dqs_oe | dqs_postamble)
         for i in range(databits//8):
             dqs_o_no_delay = Signal()
             dqs_o_delayed  = Signal()
             dqs_t          = Signal()
+            dqs_bitslip    = BitSlip(8,
+                i      = dqs_pattern.o,
+                rst    = (self._dly_sel.storage[i] & self._wdly_dq_bitslip_rst.re) | self._rst.storage,
+                slp    = self._dly_sel.storage[i] & self._wdly_dq_bitslip.re,
+                cycles = 1)
+            self.submodules += dqs_bitslip
             self.specials += Instance("OSERDESE2",
                 p_SERDES_MODE    = "MASTER",
                 p_DATA_WIDTH     = 2*nphases,
                 p_TRISTATE_WIDTH = 1,
                 p_DATA_RATE_OQ   = "DDR",
                 p_DATA_RATE_TQ   = "BUF",
-                i_RST    = ResetSignal(),
+                i_RST    = ResetSignal() | self._rst.storage,
                 i_CLK    = ClockSignal(ddr_clk) if with_odelay else ClockSignal(ddr_clk+"_dqs"),
                 i_CLKDIV = ClockSignal(),
-                i_D1     = dqs_pattern.o[0],
-                i_D2     = dqs_pattern.o[1],
-                i_D3     = dqs_pattern.o[2],
-                i_D4     = dqs_pattern.o[3],
-                i_D5     = dqs_pattern.o[4],
-                i_D6     = dqs_pattern.o[5],
-                i_D7     = dqs_pattern.o[6],
-                i_D8     = dqs_pattern.o[7],
+                **{f"i_D{n+1}": dqs_bitslip.o[n] for n in range(8)},
                 i_OCE    = 1,
                 o_OFB    = dqs_o_no_delay if with_odelay else Signal(),
                 o_OQ     = Signal() if with_odelay else dqs_o_no_delay,
                 i_TCE    = 1,
-                i_T1     = ~dqs_oe_delayed,
+                i_T1     = ~dqs_oe_delay.output,
                 o_TQ     = dqs_t,
             )
             if with_odelay:
@@ -375,39 +256,67 @@ class S7DDRPHY(Module, AutoCSR):
                     p_ODELAY_TYPE           = "VARIABLE",
                     p_ODELAY_VALUE          = half_sys8x_taps,
                     i_C        = ClockSignal(),
-                    i_LD       = self._dly_sel.storage[i] & self._wdly_dqs_rst.re,
+                    i_LD       = (self._dly_sel.storage[i] & self._wdly_dqs_rst.re) | self._rst.storage,
                     i_CE       = self._dly_sel.storage[i] & self._wdly_dqs_inc.re,
                     i_LDPIPEEN = 0,
                     i_INC      = 1,
                     o_ODATAIN  = dqs_o_no_delay,
                     o_DATAOUT  = dqs_o_delayed
                 )
-            self.specials += [
-                Instance("IDELAYE2",
-                    p_DELAY_SRC             = "IDATAIN",
-                    p_SIGNAL_PATTERN        = "DATA",
-                    p_CINVCTRL_SEL          = "FALSE",
-                    p_HIGH_PERFORMANCE_MODE = "TRUE",
-                    p_REFCLK_FREQUENCY      = iodelay_clk_freq/1e6,
-                    p_PIPE_SEL              = "FALSE",
-                    p_IDELAY_TYPE           = "FIXED",
-                    p_IDELAY_VALUE          = half_sys8x_taps,
-                    i_IDATAIN               = dqs_i[i],
-                    o_DATAOUT               = dqs_i_delayed[i]
-                )
-            ]
             self.specials += Instance("IOBUFDS",
                 i_T    = dqs_t,
                 i_I    = dqs_o_delayed if with_odelay else dqs_o_no_delay,
-                o_O    = dqs_i[i],
                 io_IO  = pads.dqs_p[i],
                 io_IOB = pads.dqs_n[i],
             )
 
+        # DM ---------------------------------------------------------------------------------------
+        if hasattr(pads, "dm"):
+            for i in range(databits//8):
+                dm_o_nodelay = Signal()
+                dm_o_bitslip = BitSlip(8,
+                    i      = Cat(*[dfi.phases[n//2].wrdata_mask[n%2*databits//8+i] for n in range(8)]),
+                    rst    = (self._dly_sel.storage[i] & self._wdly_dq_bitslip_rst.re) | self._rst.storage,
+                    slp    = self._dly_sel.storage[i] & self._wdly_dq_bitslip.re,
+                    cycles = 1)
+                self.submodules += dm_o_bitslip
+                self.specials += Instance("OSERDESE2",
+                    p_SERDES_MODE    = "MASTER",
+                    p_DATA_WIDTH     = 2*nphases,
+                    p_TRISTATE_WIDTH = 1,
+                    p_DATA_RATE_OQ   = "DDR",
+                    p_DATA_RATE_TQ   = "BUF",
+                    i_RST    = ResetSignal() | self._rst.storage,
+                    i_CLK    = ClockSignal(ddr_clk),
+                    i_CLKDIV = ClockSignal(),
+                    **{f"i_D{n+1}": dm_o_bitslip.o[n] for n in range(8)},
+                    i_OCE    = 1,
+                    o_OQ     = dm_o_nodelay if with_odelay else pads.dm[i],
+                )
+                if with_odelay:
+                    self.specials += Instance("ODELAYE2",
+                        p_SIGNAL_PATTERN        = "DATA",
+                        p_DELAY_SRC             = "ODATAIN",
+                        p_CINVCTRL_SEL          = "FALSE",
+                        p_HIGH_PERFORMANCE_MODE = "TRUE",
+                        p_PIPE_SEL              = "FALSE",
+                        p_REFCLK_FREQUENCY      = iodelay_clk_freq/1e6,
+                        p_ODELAY_TYPE           = "VARIABLE",
+                        p_ODELAY_VALUE          = 0,
+                        i_C        = ClockSignal(),
+                        i_LD       = (self._dly_sel.storage[i] & self._wdly_dq_rst.re) | self._rst.storage,
+                        i_LDPIPEEN = 0,
+                        i_CE       = self._dly_sel.storage[i] & self._wdly_dq_inc.re,
+                        i_INC      = 1,
+                        o_ODATAIN  = dm_o_nodelay,
+                        o_DATAOUT  = pads.dm[i],
+                    )
+
         # DQ ---------------------------------------------------------------------------------------
-        dq_oe         = Signal()
-        dq_oe_delayed = Signal() # Tristate control is asynchronous, needs to be delayed.
-        self.sync += dq_oe_delayed.eq(dqs_pattern.preamble | dq_oe | dqs_pattern.postamble)
+        dq_oe = Signal()
+        dq_oe_delay = TappedDelayLine(ntaps=2 if nphases == 4 else 1)
+        self.submodules += dq_oe_delay
+        self.comb += dq_oe_delay.input.eq(dqs_preamble | dq_oe | dqs_postamble)
         for i in range(databits):
             dq_o_nodelay = Signal()
             dq_o_delayed = Signal()
@@ -415,104 +324,51 @@ class S7DDRPHY(Module, AutoCSR):
             dq_i_delayed = Signal()
             dq_t         = Signal()
             dq_i_data    = Signal(8)
-            self.specials += [
-                Instance("OSERDESE2",
-                    p_SERDES_MODE    = "MASTER",
-                    p_DATA_WIDTH     = 2*nphases,
-                    p_TRISTATE_WIDTH = 1,
-                    p_DATA_RATE_OQ   = "DDR",
-                    p_DATA_RATE_TQ   = "BUF",
-                    i_RST    = ResetSignal(),
-                    i_CLK    = ClockSignal(ddr_clk),
-                    i_CLKDIV = ClockSignal(),
-                    i_D1     = dfi.phases[0].wrdata[i],
-                    i_D2     = dfi.phases[0].wrdata[databits+i],
-                    i_D3     = dfi.phases[1].wrdata[i],
-                    i_D4     = dfi.phases[1].wrdata[databits+i],
-                    i_D5     = dfi.phases[2].wrdata[i],
-                    i_D6     = dfi.phases[2].wrdata[databits+i],
-                    i_D7     = dfi.phases[3].wrdata[i],
-                    i_D8     = dfi.phases[3].wrdata[databits+i],
-                    i_TCE    = 1,
-                    i_T1     = ~dq_oe_delayed,
-                    o_TQ     = dq_t,
-                    i_OCE    = 1,
-                    o_OQ     = dq_o_nodelay,
-                )
-            ]
-            if interface_type == "NETWORKING":
-                self.specials += [
-                    Instance("ISERDESE2",
-                        p_SERDES_MODE    = "MASTER",
-                        p_INTERFACE_TYPE = "NETWORKING",
-                        p_DATA_WIDTH     = 2*nphases,
-                        p_DATA_RATE      = "DDR",
-                        p_NUM_CE         = 1,
-                        p_IOBDELAY       = "IFD",
-                        i_RST     = ResetSignal(),
-                        i_CLK     = ClockSignal(ddr_clk),
-                        i_CLKB    = ~ClockSignal(ddr_clk),
-                        i_CLKDIV  = ClockSignal(),
-                        i_BITSLIP = 0,
-                        i_CE1     = 1,
-                        i_DDLY    = dq_i_delayed,
-                        o_Q8      = dq_i_data[0],
-                        o_Q7      = dq_i_data[1],
-                        o_Q6      = dq_i_data[2],
-                        o_Q5      = dq_i_data[3],
-                        o_Q4      = dq_i_data[4],
-                        o_Q3      = dq_i_data[5],
-                        o_Q2      = dq_i_data[6],
-                        o_Q1      = dq_i_data[7],
-                    )
-                ]
-            elif interface_type == "MEMORY":
-                _dq_i_data_sys2x = Signal(4)
-                self.specials += [
-                    Instance("ISERDESE2",
-                        p_SERDES_MODE    = "MASTER",
-                        p_INTERFACE_TYPE = "MEMORY",
-                        p_DATA_WIDTH     = nphases,
-                        p_DATA_RATE      = "DDR",
-                        p_NUM_CE         = 1,
-                        p_IOBDELAY       = "IFD",
-                        i_RST     = ResetSignal(),
-                        i_CLK     = dqs_i_delayed[i//8],
-                        i_CLKB    = ~dqs_i_delayed[i//8],
-                        i_OCLK    = ClockSignal("sys4x"),
-                        i_OCLKB   = ~ClockSignal("sys4x"),
-                        i_CLKDIV  = ClockSignal("sys2x"),
-                        i_BITSLIP = 0,
-                        i_CE1     = 1,
-                        i_DDLY    = dq_i_delayed,
-                        o_Q4      = _dq_i_data_sys2x[0],
-                        o_Q3      = _dq_i_data_sys2x[1],
-                        o_Q2      = _dq_i_data_sys2x[2],
-                        o_Q1      = _dq_i_data_sys2x[3],
-                    )
-                ]
-                self.sync.sys2x += [
-                    dq_i_data[:4].eq(dq_i_data[4:]),
-                    dq_i_data[4:].eq(_dq_i_data_sys2x),
-                ]
-
-            dq_bitslip = BitSlip(8,
-                rst    = self._dly_sel.storage[i//8] & self._rdly_dq_bitslip_rst.re,
+            dq_o_bitslip = BitSlip(8,
+                i      = Cat(*[dfi.phases[n//2].wrdata[n%2*databits+i] for n in range(8)]),
+                rst    = (self._dly_sel.storage[i//8] & self._wdly_dq_bitslip_rst.re) | self._rst.storage,
+                slp    = self._dly_sel.storage[i//8] & self._wdly_dq_bitslip.re,
+                cycles = 1)
+            self.submodules += dq_o_bitslip
+            self.specials += Instance("OSERDESE2",
+                p_SERDES_MODE    = "MASTER",
+                p_DATA_WIDTH     = 2*nphases,
+                p_TRISTATE_WIDTH = 1,
+                p_DATA_RATE_OQ   = "DDR",
+                p_DATA_RATE_TQ   = "BUF",
+                i_RST    = ResetSignal() | self._rst.storage,
+                i_CLK    = ClockSignal(ddr_clk),
+                i_CLKDIV = ClockSignal(),
+                **{f"i_D{n+1}": dq_o_bitslip.o[n] for n in range(8)},
+                i_TCE    = 1,
+                i_T1     = ~dq_oe_delay.output,
+                o_TQ     = dq_t,
+                i_OCE    = 1,
+                o_OQ     = dq_o_nodelay,
+            )
+            dq_i_bitslip = BitSlip(8,
+                rst    = (self._dly_sel.storage[i//8] & self._rdly_dq_bitslip_rst.re) | self._rst.storage,
                 slp    = self._dly_sel.storage[i//8] & self._rdly_dq_bitslip.re,
-                cycles = 2)
-            self.submodules += dq_bitslip
-            self.comb += dq_bitslip.i.eq(dq_i_data)
-            self.comb += [
-                dfi.phases[0].rddata[i].eq(dq_bitslip.o[0]),
-                dfi.phases[1].rddata[i].eq(dq_bitslip.o[2]),
-                dfi.phases[2].rddata[i].eq(dq_bitslip.o[4]),
-                dfi.phases[3].rddata[i].eq(dq_bitslip.o[6]),
-                dfi.phases[0].rddata[databits+i].eq(dq_bitslip.o[1]),
-                dfi.phases[1].rddata[databits+i].eq(dq_bitslip.o[3]),
-                dfi.phases[2].rddata[databits+i].eq(dq_bitslip.o[5]),
-                dfi.phases[3].rddata[databits+i].eq(dq_bitslip.o[7])
-            ]
-
+                cycles = 1)
+            self.submodules += dq_i_bitslip
+            self.specials += Instance("ISERDESE2",
+                p_SERDES_MODE    = "MASTER",
+                p_INTERFACE_TYPE = "NETWORKING",
+                p_DATA_WIDTH     = 2*nphases,
+                p_DATA_RATE      = "DDR",
+                p_NUM_CE         = 1,
+                p_IOBDELAY       = "IFD",
+                i_RST     = ResetSignal() | self._rst.storage,
+                i_CLK     = ClockSignal(ddr_clk),
+                i_CLKB    = ~ClockSignal(ddr_clk),
+                i_CLKDIV  = ClockSignal(),
+                i_BITSLIP = 0,
+                i_CE1     = 1,
+                i_DDLY    = dq_i_delayed,
+                **{f"o_Q{n+1}": dq_i_bitslip.i[8-1-n] for n in range(8)},
+            )
+            for n in range(8):
+                self.comb += dfi.phases[n//2].rddata[n%2*databits+i].eq(dq_i_bitslip.o[n])
             if with_odelay:
                 self.specials += Instance("ODELAYE2",
                     p_SIGNAL_PATTERN        = "DATA",
@@ -524,67 +380,71 @@ class S7DDRPHY(Module, AutoCSR):
                     p_ODELAY_TYPE           = "VARIABLE",
                     p_ODELAY_VALUE          = 0,
                     i_C        = ClockSignal(),
-                    i_LD       = self._dly_sel.storage[i//8] & self._wdly_dq_rst.re,
+                    i_LD       = (self._dly_sel.storage[i//8] & self._wdly_dq_rst.re)| self._rst.storage,
                     i_LDPIPEEN = 0,
                     i_CE       = self._dly_sel.storage[i//8] & self._wdly_dq_inc.re,
                     i_INC      = 1,
                     o_ODATAIN  = dq_o_nodelay,
                     o_DATAOUT  = dq_o_delayed,
                 )
-            self.specials += [
-                Instance("IDELAYE2",
-                    p_SIGNAL_PATTERN        = "DATA",
-                    p_DELAY_SRC             = "IDATAIN",
-                    p_CINVCTRL_SEL          = "FALSE",
-                    p_HIGH_PERFORMANCE_MODE = "TRUE",
-                    p_REFCLK_FREQUENCY      = iodelay_clk_freq/1e6,
-                    p_PIPE_SEL              = "FALSE",
-                    p_IDELAY_TYPE           = "VARIABLE",
-                    p_IDELAY_VALUE          = 0,
-                    i_C        = ClockSignal(),
-                    i_LD       = self._dly_sel.storage[i//8] & self._rdly_dq_rst.re,
-                    i_LDPIPEEN = 0,
-                    i_CE       = self._dly_sel.storage[i//8] & self._rdly_dq_inc.re,
-                    i_INC      = 1,
-                    i_IDATAIN  = dq_i_nodelay,
-                    o_DATAOUT  = dq_i_delayed
-                ),
-                Instance("IOBUF",
-                    i_I   = dq_o_delayed if with_odelay else dq_o_nodelay,
-                    o_O   = dq_i_nodelay,
-                    i_T   = dq_t,
-                    io_IO = pads.dq[i]
-                )
-            ]
+            self.specials += Instance("IDELAYE2",
+                p_SIGNAL_PATTERN        = "DATA",
+                p_DELAY_SRC             = "IDATAIN",
+                p_CINVCTRL_SEL          = "FALSE",
+                p_HIGH_PERFORMANCE_MODE = "TRUE",
+                p_REFCLK_FREQUENCY      = iodelay_clk_freq/1e6,
+                p_PIPE_SEL              = "FALSE",
+                p_IDELAY_TYPE           = "VARIABLE",
+                p_IDELAY_VALUE          = 0,
+                i_C        = ClockSignal(),
+                i_LD       = (self._dly_sel.storage[i//8] & self._rdly_dq_rst.re) | self._rst.storage,
+                i_LDPIPEEN = 0,
+                i_CE       = self._dly_sel.storage[i//8] & self._rdly_dq_inc.re,
+                i_INC      = 1,
+                i_IDATAIN  = dq_i_nodelay,
+                o_DATAOUT  = dq_i_delayed
+            )
+            self.specials += Instance("IOBUF",
+                i_I   = dq_o_delayed if with_odelay else dq_o_nodelay,
+                o_O   = dq_i_nodelay,
+                i_T   = dq_t,
+                io_IO = pads.dq[i]
+            )
 
         # Read Control Path ------------------------------------------------------------------------
-        # Creates a shift register of read commands coming from the DFI interface. This shift register
-        # is used to indicate to the DFI interface that the read data is valid.
+        # Creates a delay line of read commands coming from the DFI interface. The output is used to
+        # signal a valid read data to the DFI interface.
         #
         # The read data valid is asserted for 1 sys_clk cycle when the data is available on the DFI
         # interface, the latency is the sum of the OSERDESE2, CAS, ISERDESE2 and Bitslip latencies.
-        rddata_en      = Signal(self.settings.read_latency)
-        rddata_en_last = Signal.like(rddata_en)
-        self.comb += rddata_en.eq(Cat(dfi.phases[self.settings.rdphase].rddata_en, rddata_en_last))
-        self.sync += rddata_en_last.eq(rddata_en)
-        self.sync += [phase.rddata_valid.eq(rddata_en[-1] | self._wlevel_en.storage) for phase in dfi.phases]
+        rddata_en = TappedDelayLine(
+            signal = reduce(or_, [dfi.phases[i].rddata_en for i in range(nphases)]),
+            ntaps  = self.settings.read_latency
+        )
+        self.submodules += rddata_en
+
+        self.comb += [phase.rddata_valid.eq(rddata_en.output | self._wlevel_en.storage) for phase in dfi.phases]
 
         # Write Control Path -----------------------------------------------------------------------
-        # Creates a shift register of write commands coming from the DFI interface. This shift register
-        # is used to control DQ/DQS tristates.
-        wrdata_en = Signal(cwl_sys_latency + 2)
-        wrdata_en_last = Signal.like(wrdata_en)
-        self.comb += wrdata_en.eq(Cat(dfi.phases[self.settings.wrphase].wrdata_en, wrdata_en_last))
-        self.sync += wrdata_en_last.eq(wrdata_en)
-        self.comb += dq_oe.eq(wrdata_en[cwl_sys_latency])
+        wrtap = cwl_sys_latency - 1
+
+        # Create a delay line of write commands coming from the DFI interface. This taps are used to
+        # control DQ/DQS tristates.
+        wrdata_en = TappedDelayLine(
+            signal = reduce(or_, [dfi.phases[i].wrdata_en for i in range(nphases)]),
+            ntaps  = wrtap + 2
+        )
+        self.submodules += wrdata_en
+
+        self.comb += dq_oe.eq(wrdata_en.taps[wrtap])
         self.comb += If(self._wlevel_en.storage, dqs_oe.eq(1)).Else(dqs_oe.eq(dq_oe))
 
         # Write DQS Postamble/Preamble Control Path ------------------------------------------------
         # Generates DQS Preamble 1 cycle before the first write and Postamble 1 cycle after the last
         # write. During writes, DQS tristate is configured as output for at least 3 sys_clk cycles:
         # 1 for Preamble, 1 for the Write and 1 for the Postamble.
-        self.comb += dqs_pattern.preamble.eq( wrdata_en[cwl_sys_latency - 1]  & ~wrdata_en[cwl_sys_latency])
-        self.comb += dqs_pattern.postamble.eq(wrdata_en[cwl_sys_latency + 1]  & ~wrdata_en[cwl_sys_latency])
+        self.comb += dqs_preamble.eq( wrdata_en.taps[wrtap - 1]  & ~wrdata_en.taps[wrtap + 0])
+        self.comb += dqs_postamble.eq(wrdata_en.taps[wrtap + 1]  & ~wrdata_en.taps[wrtap + 0])
 
 # Xilinx Virtex7 (S7DDRPHY with odelay) ------------------------------------------------------------
 
